'
' RETROGRA.BI
'
' "RETROGRA" is a graphics library for "retro" graphics, precisely spoken:
' limiting to 640x480 2D graphics with 256 colours in fullscreen and windowed
' mode, which is in fact more than the original VGA standard resp. beyond the
' capabilities of homecomputers
'
' replaces SCREENSIZE, SCREENLOCK and so on, "double-buffering" is forced
'
' It features further a graphical coloured console, video terminal (VT) style
' but A.O.T. with UTF-8 support
' brings a predefined palette based upon the console standard
'
' ment for retro-style games, also text-console-based, and other "simple"
' applications, a kind of "DOSBox" without the limitations of DOS
'
' This low resolution makes it much easier to create sprites and other graphics
'
' scales the output to fit better on high resoluted monitors
'
' Works with most ordinary FreeBASIC code, but do never use SCREENLOCK etc.
' instead of the rg versions
'
' ******************************************************************************
'
' rgSCREENNEW should create a new double-buffered canvas - 640x480 8 bpp
'   use 'rgSCREENNEW WINDOWED' to not start in fullscreen mode
'
' rgSCREENDELETE "destroys" the canvas somehow, like when exiting the program
'   => unnecessary
'
' rgSCREENSHOW - call this to make all of your graphic operations visible,
'   draws mouse pointer above all
'   waits for VSynch; use it instead of SCREENLOCK, SCREENUNLOCK and SCREENCOPY
'
' rgSHIFTPALETTE (int r, int g, int b)
'   shifts the colours of the default palette by RGB values of +/- 100(%)
'   0% restores the default palette values
'   only affects colour indicies 0-215, 216-239 may be changed "manually"
'   rbg+100 makes all white rbg-100 all black
'
' rgSETSCALING (int scaling)
'   sets video scaling factor, 0 as 1 are disabling software-scaling
'
' rgSETMOUSE (int look, int x, int y)
'   can move the mouse cursor, selects a look, 0 for hide
'
' rgPRINT (string expression)
'   text console output - regard the control codes, especially ESC
'
' rgINPUT (string prompt, string var[, int limit=72])
'   primitive console input, only one-lined, always blocked at the page's end
'   overwrites background like rgPRINT, doesn't restore it
'   only for strings, no validation check => write your own input function
'
' AT(int column, int row) - creates an ESC P sequence for placing the cursor
'   using in rgPRINT expressions (like LOCATE in original FreeBASIC/PRINT)
'
' PRINTCOLOR(uint8 fgcolour, uint8 bgcolour) - returns an ESC p sequence
'   to set the printing colours
'
' CRLF - the string "\n", maybe easier to type for some
'
' Annotation: there are undocumented subs - those are internal and shouldn't
'             be used, for example all "rgc..."-SUBs (=> RetroGra-Console), but
'             not only
'
' Future enhancements:
' rgMESSAGEBOX - GUI message box
' rgFILESELECT - GUI file selector dialogue
' rgDRAWTEXT(int x, int y, int size, string text[, colour[, decorations]])
'              - graphical output of single-lined UTF-8 text,
'              multicoloured / 216 colours, "all" sizes,
'              monospaced or proportional, scaled with antialiasing,
'              selectable bitmap- or "true type" fonts
'              does not execute control codes
'              size is always height of the line in pixels, character width
'              depends on this size value
'              colour is only working for monochrome fonts
' rgSETFONT(string) - selects a font (file)
' rgGETTEXTLENGTH(int size, string text) - returns width of a text without
'              putting it out
'
' rgDRAW(x, y, colour[, count]) - draws a point or polygone line supporting
'              antialiasing and transparency(alpha)
' rgPOLYGONE(x(), y(), colour[, count]) - draws a filled polygone
' rgLINE(x1,y1,x2,y2,colour,...) - draws one line with "styles"
' rgBOX(x1,y1,x2,y2,colour,...) - draws a box line
' rgRBOX - box with rounded edges
' rgFBOX - filled box
' rgRFBOX - filled rbox
' rgELLIPSE
' rgFELLIPSE
'
' ******************************************************************************
'
' GRAPHICAL COLOURED CONSOLE
'
' the console output is based on historic terminals, where text output was
' exchangeable between monitors and printers => you might print also text
' on a fictive colour printer this way
' It's further capable of creating coloured character-based graphics like
' used by ZX-81 and C64 or "teletext"/"ceefax". Although limited, it
' outperforms clearly "ASCII art", which only uses ASCII characters to form
' graphics and mostly black-white
'
' It's part of a pseudo-standard you may call "VT-2000", derived from
' "VT-52". The 8-bit colour palette could be translated also in a greyscale,
' 4 bit and monochrome, based on "nearest colour". Many text decorations and
' features might not be supported and ignored, normally returning an error
' code or failed flag. Supports UTF-8.
'
' The font is monospaced with a fixed count of characters per line and lines
' per screen; all characters have the same size except some double sized
' characters like CJK characters, occupying 2 cells per char. Scrolls
' vertically by losing the screen content.
' In Retrogra limited to 80 characters per 30 lines on the screen only,
' therefore only 8x16 pixel sized cells and no space for many text decorations
' and looking REALLY "retro"...
'
' Each character/field has only a foreground and a background colour, but
' individually out of 256 colours simultanously
'
' Control Codes:
' Like ASCII, Codes below 0x20/32 (Space) have special functions and don't
' put out characters. Most important 27 - ESC, "Escape" also '\e': this is
' a code for very complex so called "escape sequences", which might cause
' most trouble if appearing unintended in strings
' The Nullbyte marks the end of a string, therefore stops the string processing
' even if it doesn't demand a strings' length
' The control codes aren't following the ANSI ASCII standard further -
' most important difference is Code 8 instead of 9 for the 8-character-wide
' tabulator
' although 13/0x0D/'\n' works as return/line break, the outdated sequence 13+10
' does not => this equals 11+2 instead ("carriage return" + "line feed")
' so use '\n' for line breaks
' use code 0x1A/26 to access the control characters' graphic, followed by
' the about 32 higher character code
' all multi-character codes have to be together in the same string; you can't
' send "parameter bytes" or data later with the next PRINT call
'
' Note, that all characters above code 127 have to be UTF-8 encoded always,
' a Nullbyte always ends the string/output everywhere.

DECLARE SUB rgINITMOUSE (BYVAL AS STRING="")
DECLARE SUB rgINITCONSOLE
DECLARE SUB rgSCREENSHOW

DIM SHARED AS BOOLEAN debugoutinitialised=FALSE
SUB DEBUGOUT(BYVAL text AS string) ' primitive debug utility function
  DIM AS INTEGER fn=FREEFILE
  IF debugoutinitialised THEN
    OPEN "debug.txt" FOR APPEND AS #fn
  ELSE
    OPEN "debug.txt" FOR OUTPUT AS #fn
  ENDIF
  PRINT #fn, text
  CLOSE #fn
  debugoutinitialised=TRUE
END SUB

#INCLUDE "fbgfx.bi"

DIM SHARED AS LONG retrograpalette(256)
DIM SHARED AS INTEGER retrograscaling, retrogra_scrstartx, retrogra_scrstarty, retrogra_scalestartx, retrogra_scalestarty
DIM SHARED AS DOUBLE retrogralastrefresh

DIM SHARED AS ANY PTR retrograscreenupdatethread
DIM SHARED AS boolean retrograstopthreads=FALSE

' rgSCREENNEW canvas flags
CONST AS INTEGER FULLSCREEN=1
CONST AS INTEGER WINDOWED=0
CONST AS INTEGER FORCEDFULLSCREEN=3
CONST AS INTEGER FRAMELESS=4
CONST AS INTEGER UNSCALED=8

' mouse cursor stuff -->
DIM SHARED AS INTEGER retrogramousepointer
DIM SHARED AS BYTE PTR retrogramousepointerbitmap(16)
DIM SHARED AS BYTE PTR retrogramousepointermask(16)
DIM SHARED AS BYTE PTR retrogramousebackbuffer ' buffers background of mouse pointer
DIM SHARED AS INTEGER retrogramousepointerxo(16),retrogramousepointeryo(16)

' rgSETMOUSE form names
CONST AS INTEGER rgDISABLE=0
CONST AS INTEGER rgARROW=1
CONST AS INTEGER rgBRACKET=2
CONST AS INTEGER rgBUSY=3
CONST AS INTEGER rgFAINTCROSS=4
CONST AS INTEGER rgCROSS=5
CONST AS INTEGER rgCROSSHAIR=6
CONST AS INTEGER rgPOINTING=7
CONST AS INTEGER rgDRAGGING=8
CONST AS INTEGER rgQUESTION=9
CONST AS INTEGER rgBLOCKED=10

' graphical console stuff -->
DIM SHARED AS UBYTE rgcfont() ' buffer containing font data
DIM SHARED AS UInteger rgcfontpageoffset(512) ' offsets to Unicode "pages" 0x0000+, 0x0080+, 0x0100+, 0x0180+ ...
DIM SHARED AS UInteger rgcfontlinesize(512) ' offset/size of a line in the corr. "page"
DIM SHARED AS USHORT rgcflags ' console mode flags
'                     1 = vertical overflow - scrolling on (default)
'                     2 = line overflow on (default)
'                    32 = always visible (maybe blinking) cursor
'                    64 = reserved / no effect, see '192'
'                   128 = sound on (default) - only BEL(alert) when line overflow is blocked
'                   192 = full sound enviroment (typing sound on inputs, carriage return...)
'                         note: the respective control codes are putting out sounds independently
DIM SHARED AS INTEGER rgccurx, rgccury ' current console cursor position 0-79 / 0-29
DIM SHARED AS UBYTE rgcfgcolour, rgcbgcolour ' current console char foreground+background colour
DIM SHARED AS USHORT rgccharflags ' current console char flags; decorations a.o.t.
'                     1 = "not empty" / normal
'                     2 = bold
'                     4 = italic
'                     8 = underlined
'                    16 = cancelled
'                    32 = "disabled" / "grey"
'                    64 = outlined
'                   128 = inverted
'                   512 = superscript
'                   256 = subscript
'                  1024 = double width (occupying double of cells)
'                  2048 = double height
'                  4096 = compressed (only for proportional outputs)
'   16384, 32768, 49152 = "double char" (right, below, right+below)
TYPE rgcCELL
  flags AS USHORT
  fgcolour AS UBYTE
  bgcolour AS UBYTE
  char AS ULONG
END TYPE
DIM SHARED AS rgcCELL rgccell(80,30)

' the "function" CRLF... (line break)
#DEFINE CRLF CHR(13)

SUB rgSCREENUPDATETHREAD(BYVAL parameter AS ANY PTR)
  DO
    IF (TIMER-retrogralastrefresh)>0.5 THEN
      rgSCREENSHOW
    ENDIF
    SLEEP 50,1
  LOOP WHILE retrograstopthreads=FALSE
END SUB

SUB rgSCREENNEW (BYVAL flags AS INTEGER=FULLSCREEN, BYVAL scaling AS INTEGER=1)
  ' flags: WINDOWED <-> FULLSCREEN,
  '        "FORCEDFULLSCREEN" creates a desktop sized canvas for fullscreen
  '        FRAMELESS = borderless window
  '        UNSCALED = perform no scaling => 640x480 window content
  ' scaling: forced scaling factor, 0 = no scaling, 1 = auto(default)
  '          if windowed mode, window size = scaling size, must not exceed monitor resolution
  DIM AS INTEGER c, r, g, b, dc

  ' *** creating colour console palette ***
  ' assigning RGB-scale colours 0-215
  ' in FreeBASIC format => 6 bits, 0-63 per colour channel only
  c=0
  FOR b=0 TO 63 STEP 12
    IF b=24 THEN b=27
    FOR g=0 TO 63 STEP 12
      IF g=24 THEN g=27
      FOR r=0 TO 63 STEP 12
        IF r=24 THEN r=27
        retrograpalette(c)=(b SHL 16) + (g SHL 8) + r
        c+=1
      NEXT
    NEXT
  NEXT
  ' free/user defined colours => dummy colour values
  FOR c=216 TO 238 STEP 2
    retrograpalette(c)=&H102030
    retrograpalette(c+1)=&H112233
  NEXT
  ' 0xF0-0xFF console special effect colours
  retrograpalette(240)=0 ' "permanent" black (turns to 0 in memory)
  ' also dummy colours
  FOR c=241 TO 254
    retrograpalette(c)=(((13*(c MOD 2))+50) SHL 16)+(13*(c MOD 2))+50
  NEXT
  retrograpalette(255)=&H303F3F ' inverting effect, dummy colour

  ' *** creating canvas ***
  DIM AS INTEGER res, w, h, d
  IF (flags AND FORCEDFULLSCREEN) THEN ' force desktop sized fullscreen 8 bpp with scaling
    res=0 ' don't try 640x480 in fullscreen
  ELSE
    res=SCREENLIST(8) ' list of 8 bpp fullscreen graphic modes
    WHILE res
      w=HIWORD(res)
      h=LOWORD(res)
      IF (w=640 AND h=480) THEN EXIT WHILE
      res=SCREENLIST
    WEND
  ENDIF
  IF res THEN ' fullscreen available for 640x480 8 bpp
    retrograscaling=0 ' scaling off
    IF (flags AND FULLSCREEN) THEN
      SCREENRES 640, 480, 8, 3, FB.GFX_FULLSCREEN + FB.GFX_NO_FRAME
    ENDIF
  ELSE ' 640x480 8 bpp not supported
    ' creating a desktop sized canvas for fullscreen
    '   or a window best scaled to desktop size
    SCREENINFO w, h, d ' get desktop resolution
    IF scaling THEN
      retrogra_scrstartx=0 ' start coords in virtual(640x480) screen
      retrogra_scrstarty=0
      retrograscaling=h/480 ' most likely, height doesn't fit as often in desktop resolutions like width do
      IF (w/640)<retrograscaling THEN retrograscaling=w/640 ' but otherwise...
      IF scaling<=retrograscaling THEN
        IF scaling>1 THEN ' not auto scaling => user demand
          retrograscaling=scaling
        ENDIF
      ELSE ' scaling exceeds screen size
        ' determine whether cutting loss is small
        IF (480-INT(h/scaling)<33 AND 640-INT(w/scaling)<33) THEN ' cut only up to 32 pixels
          retrograscaling=scaling
          retrogra_scrstartx=(640-INT(w/retrograscaling)) SHR 1 ' start copying original from these coords
          IF retrogra_scrstartx<1 THEN retrogra_scrstartx=0
          retrogra_scrstarty=(480-INT(h/retrograscaling)) SHR 1
          IF retrogra_scrstarty<1 THEN retrogra_scrstarty=0
        ENDIF ' ELSE keep the previously calculated value
      ENDIF
      ' determine offsets
      IF NOT (retrogra_scrstartx OR retrogra_scrstarty) THEN ' if this offsets>0 => starting at 0
        retrogra_scalestartx=(w-640*retrograscaling) SHR 1 ' centers output
        retrogra_scalestarty=(h-480*retrograscaling) SHR 1
      ENDIF
    ENDIF

    IF (flags AND FULLSCREEN) THEN ' fullscreen modes
      SCREENRES w, h, 8, 3, FB.GFX_FULLSCREEN + FB.GFX_NO_FRAME
      ' #todo: error 8 bpp failed / no canvas
    ELSE ' windowed modes
      w=640*retrograscaling
      h=480*retrograscaling
      IF (flags AND FRAMELESS) THEN ' window without decorations
        SCREENRES w, h, 8, 3, FB.GFX_NO_FRAME
      ELSE
        SCREENRES w, h, 8, 3
      ENDIF
    ENDIF
  ENDIF

  SCREENCOPY 2,0 ' 
  SCREENSET 2,0
  ' *** load mouse pointer graphics ***
  rgINITMOUSE

  ' *** init console / load console font ***
  rgINITCONSOLE

  PALETTE USING retrograpalette

  ' start refresh-forcing thread
  retrograstopthreads=FALSE
  retrogralastrefresh=TIMER
  retrograscreenupdatethread=THREADCREATE(@rgSCREENUPDATETHREAD)
END SUB

SUB rgSCREENDELETE
  ' "deconstructor"

  'stop all threads
  retrograstopthreads=TRUE
  THREADWAIT(retrograscreenupdatethread)

  ' free buffers
  IMAGEDESTROY retrogramousebackbuffer
  FOR i AS INTEGER=0 TO 15
    IMAGEDESTROY retrogramousepointerbitmap(i)
    IMAGEDESTROY retrogramousepointermask(i)
  NEXT

  ' delete screens
  SCREEN 0
END SUB

SUB rgSCREENSHOW
  DIM AS INTEGER x, y, o, scw, mx, my
  DIM AS BYTE c
  DIM AS BYTE PTR vga_screen, logical_screen

  IF (TIMER-retrogralastrefresh<0.005) THEN EXIT SUB ' avoid "ridiculous refresh rates"

  ' place mousecursor (if active)
  IF retrogramousepointer THEN
    SCREENSET 2,0
    GETMOUSE(mx, my)
    IF retrograscaling THEN
      mx/=retrograscaling
      my/=retrograscaling
      IF mx>639 THEN mx=639
      IF my>479 THEN my=479
      mx-=retrogramousepointerxo(retrogramousepointer-1)
      my-=retrogramousepointeryo(retrogramousepointer-1)
      IF (mx<0 OR my<0) THEN
        IF (mx<0 AND my<0) THEN
          GET (0,0)-(15,15),retrogramousebackbuffer
        ELSE
          IF mx<0 THEN
            GET (0,my)-(15,my+15),retrogramousebackbuffer
          ELSE
            GET (mx,0)-(mx+15,15),retrogramousebackbuffer
          ENDIF
        ENDIF
      ELSE
        GET (mx,my)-(mx+15,my+15),retrogramousebackbuffer
      ENDIF
      PUT (mx, my), retrogramousepointermask(retrogramousepointer-1),AND
      PUT (mx, my), retrogramousepointerbitmap(retrogramousepointer-1),OR
    ELSE
      IF mx>639 THEN mx=639
      IF my>479 THEN my=479
      GET (mx,my)-(mx+15,my+15),retrogramousebackbuffer
      PUT (mx, my), retrogramousepointermask(retrogramousepointer-1),AND
      PUT (mx, my), retrogramousepointerbitmap(retrogramousepointer-1),OR
    ENDIF
  ENDIF

  ' check whether scaling is done / perform scaling
  IF retrograscaling THEN
    ' getting video buffers, scaling-copy memory to memory
    vga_screen=SCREENPTR
    SCREENSET 1,0
    logical_screen=SCREENPTR
    SCREENINFO scw,y ' getting screen width
    SELECT CASE retrograscaling
      CASE 2:
        ' scaling 2 - 2x2 copying
        FOR y=0 TO 479
          FOR x=0 TO 639
            c=*(vga_screen+y*scw+x)
            o=((y SHL 1)+retrogra_scalestarty)*scw+(x SHL 1)+retrogra_scalestartx
            *(logical_screen+o)=c
            o+=1
            *(logical_screen+o)=c
            o+=scw
            o-=1
            *(logical_screen+o)=c
            o+=1
            *(logical_screen+o)=c
          NEXT
        NEXT
        SCREENSYNC
        SCREENCOPY 1,0
      CASE ELSE:
        ' scaling from 3 above - 3x3, 4x4 etc. copies
        DIM AS INTEGER x2, y2, screndx, screndy
        screndx=639-retrogra_scrstartx
        screndy=479-retrogra_scrstarty
        FOR y=retrogra_scrstarty TO screndy
          FOR x=retrogra_scrstartx TO screndx
            c=*(vga_screen+y*scw+x)
            o=((y*retrograscaling)+retrogra_scalestarty)*scw+(x*retrograscaling)+retrogra_scalestartx
            y2=retrograscaling
            DO
              x2=retrograscaling
              DO
                *(logical_screen+o)=c
                o+=1
                x2-=1
              LOOP WHILE x2
              o+=scw
              o-=retrograscaling
              y2-=1
            LOOP WHILE y2
          NEXT
        NEXT
        SCREENSYNC
        SCREENCOPY 1,0 ' copy scaled to physical screen
    END SELECT
  ELSE
    SCREENSYNC
    SCREENCOPY 2,0 ' copy virtual to physical screen
  ENDIF
  SCREENSET 2,0 ' output in virtual screen
  retrogralastrefresh=TIMER

  ' remove mousecursor "sprite" (if active)
  IF retrogramousepointer THEN
    IF (mx<0) THEN mx=0
    IF (my<0) THEN my=0
    PUT (mx,my),retrogramousebackbuffer,PSET
  ENDIF
END SUB

SUB rgSETSCALING (BYVAL scaling AS INTEGER)
  IF scaling<1 THEN
    retrograscaling=0 ' scaling disabled
    EXIT SUB
  ENDIF
  IF scaling=retrograscaling THEN EXIT SUB
  ' check scaling to be valid, adapt scaling variables
END SUB

SUB rgSHIFTPALETTE (BYVAL rperc AS BYTE=0, BYVAL gperc AS BYTE=0, BYVAL bperc AS BYTE=0)
  DIM AS ULONG i, r, g, b
  DIM AS LONG newpal(256)

  FOR i=0 TO 215
    IF rperc=0 THEN ' keep value
      r=retrograpalette(i) AND &H3F
    ELSEIF rperc=100 THEN ' max. value (of white)
      r=&H3F
    ELSEIF rperc=-100 THEN ' min. value (of black)
      r=0
    ELSE
      r=retrograpalette(i) AND &H3F
      IF rperc>0 THEN
        r+=(63-r)*rperc/100
      ELSE
        r-=r*rperc/-100
      ENDIF
      r AND=63
    ENDIF
    IF gperc=0 THEN ' keep value
      g=retrograpalette(i) AND &H3F00
    ELSEIF gperc=100 THEN ' max. value (of white)
      g=&H3F00
    ELSEIF gperc=-100 THEN ' min. value (of black)
      g=0
    ELSE
      g=retrograpalette(i) AND &H3F00
      IF gperc>0 THEN
        g+=(&H3F00-g)*gperc/100
      ELSE
        g-=g*gperc/-100
      ENDIF
      g AND=&H3F00
    ENDIF
    IF bperc=0 THEN ' keep value
      b=retrograpalette(i) AND &H3F0000
    ELSEIF bperc=100 THEN ' max. value (of white)
      b=&H3F0000
    ELSEIF bperc=-100 THEN ' min. value (of black)
      b=0
    ELSE
      b=retrograpalette(i) AND &H3F0000
      IF bperc>0 THEN
        b+=(&H3F0000-b)*bperc/100
      ELSE
        b-=b*bperc/-100
      ENDIF
      b AND=&H3F0000
    ENDIF
    newpal(i)=r+g+b
  NEXT
  FOR i=216 TO 254
    newpal(i)=retrograpalette(i)
  NEXT
  PALETTE USING newpal
END SUB

SUB rgINITMOUSE (BYVAL filename AS STRING="")
  IF retrogramousepointerbitmap(0) THEN EXIT SUB ' already initialised
  DIM AS INTEGER i, x, y, c
  DIM mousepointers AS ANY PTR
  ' init buffer for pointer graphics - always 8 pointers 16x16
  mousepointers=IMAGECREATE(512,16)
  ' load pointer graphics BMP
  IF filename="" THEN ' no filename => load defaults
    i=BLOAD("retrogra.rsc/mouse.bmp",mousepointers,)
  ELSE
    i=BLOAD(filename,mousepointers,)
    IF i THEN ' likely given file not found => load defaults
      i=BLOAD("retrogra.rsc/mouse.bmp",mousepointers,)
    ENDIF
  ENDIF
  retrogramousebackbuffer=IMAGECREATE(16,16) ' buffers background of mouse pointer
  ' create sprites from image, get hotspots ...
  FOR i=0 TO 15
    retrogramousepointerbitmap(i)=IMAGECREATE(16,16)
    retrogramousepointermask(i)=IMAGECREATE(16,16)
    GET mousepointers,(i*32,0)-(i*32+15,15),retrogramousepointerbitmap(i)
    GET mousepointers,(i*32+16,0)-(i*32+31,15),retrogramousepointermask(i)
    retrogramousepointerxo(i)=1
    retrogramousepointeryo(i)=1
    FOR y=0 TO 15
      FOR x=0 TO 15
        c=POINT(x,y,retrogramousepointermask(i))
        IF c=5 THEN ' hotspot opaque
          retrogramousepointerxo(i)=x
          retrogramousepointeryo(i)=y
          PSET retrogramousepointermask(i), (x, y), 0
          EXIT FOR, FOR
        ENDIF
        IF c=30 THEN ' hotspot transparent
          retrogramousepointerxo(i)=x
          retrogramousepointeryo(i)=y
          PSET retrogramousepointermask(i), (x, y), 255
          EXIT FOR, FOR
        ENDIF
      NEXT
    NEXT
  NEXT
  IMAGEDESTROY mousepointers
  SETMOUSE , , 0
  retrogramousepointer=1
END SUB

SUB rgINITCONSOLE
  DIM AS INTEGER i, x, y
  DIM AS INTEGER fn
  DIM AS BYTE PTR address

  ' load console font file data into memory
  fn=FREEFILE
  IF OPEN("retrogra.rsc/c16x8.fnt" FOR BINARY AS #fn)=0 THEN
    REDIM rgcfont(LOF(fn)-1) ' buffer containing font data
    GET #fn, 1, rgcfont() ' simply load the whole file
    CLOSE #fn

    ' get Unicode "page" offsets
    address=@rgcfont(0)+&H80
    FOR i=0 TO 511
      rgcfontpageoffset(i)=PEEK(ULONG, address)
      address+=4
    NEXT
    ' get font data line width of each "page" / set offset to first char data
    FOR i=0 TO 511
      IF rgcfontpageoffset(i) THEN
        '  address=@rgcfont(0)+rgcfontpageoffset(i)
        rgcfontlinesize(i)=128 ' #todo: read data really, but FNT file not yet ready !
        rgcfontpageoffset(i)+=4
      ENDIF
    NEXT
  ELSE
    ' #todo: file open error => fatal
    rgcfontpageoffset(0)=0 ' definitely wrong value
  ENDIF

  ' set default/starting values
  rgccurx=0
  rgccury=0
  rgcfgcolour=0
  rgcbgcolour=215
  rgcflags=131 ' default: overflow+scrolling+BEL sound
  rgccharflags=1 ' default: normal / draw character

  ' "clear" "screen"
  ' default is black on white
  ' if it should start e.g. black, you have to overdraw the screen
  ' before the first rgSCREENSHOW call
  FOR y=0 TO 29
    FOR x=0 TO 79
      rgccell(x,y).flags=0
      rgccell(x,y).fgcolour=0 ' black
      rgccell(x,y).bgcolour=215 ' white
      rgccell(x,y).char=0
    NEXT
  NEXT
  LINE (0,0)-(639,479),215,BF
END SUB

SUB rgGETMOUSE (BYREF mx AS INTEGER, BYREF my AS INTEGER, BYREF wheel AS INTEGER=0, BYREF button AS INTEGER=0)
  GETMOUSE(mx, my, wheel, button)
  IF retrograscaling THEN
    mx/=retrograscaling
    my/=retrograscaling
  ENDIF
  IF mx>639 THEN mx=639
  IF my>479 THEN my=479
END SUB

SUB rgSETMOUSE (BYVAL form AS INTEGER=-1, BYVAL x AS INTEGER=-1, BYVAL y AS INTEGER=-1)
  ' form: -1 illegal / don't change - 0-16 allowed
  '        0 invisible / disable
  '       standards:
  '        1 arrow / pointer
  '        2 doublebracket
  '        3 busy
  '        4 thin cross
  '        5 thick cross
  '        6 "crosshair"
  '        7 pointing hand
  '        8 dragging hand
  '        9 question mark
  '       10 blocked
  IF (form>=0 AND form<=16) THEN retrogramousepointer=form
  IF (x<>-1 OR y<>-1) THEN
    IF retrograscaling THEN
      IF x<>-1 THEN x*=retrograscaling
      IF y<>-1 THEN y*=retrograscaling
    ENDIF
    SETMOUSE x, y, 0
  ENDIF
END SUB

FUNCTION AT(BYVAL column AS byte, BYVAL row AS byte) AS STRING
  IF column>0 THEN
    IF column>80 THEN column=80 ' outside console area
  ELSE
    column=1 ' 0/negative are illegal values - no Nullbytes allowed also
  ENDIF
  IF row>0 THEN
    IF row>30 THEN row=30 ' outside console area
  ELSE
    row=1 ' 0/negative are illegal values - no Nullbytes allowed also
  ENDIF
  AT=CHR(27)+"P"+CHR(column,row)
END FUNCTION

FUNCTION PRINTCOLOR(BYVAL fgcolour AS UBYTE, BYVAL bgcolour AS UBYTE) AS STRING
  DIM AS STRING f, b
  DIM AS INTEGER i

  IF fgcolour THEN
    IF fgcolour>127 THEN ' requires UTF-8 encoding
      f=CHR(((fgcolour SHR 6) OR 192), ((fgcolour AND 63)+128)) ' only 8 bit; 0-255 possible
    ELSE
      f=CHR(fgcolour)
    ENDIF
  ELSE ' colour 0 => translate to "permanent black" 0xF0/240 (also in UTF-8)
    f=CHR(195, 144)
  ENDIF
  IF bgcolour THEN
    IF bgcolour>127 THEN ' requires UTF-8 encoding
      b=CHR(((bgcolour SHR 6) OR 192), ((bgcolour AND 63)+128)) ' only 8 bit; 0-255 possible
    ELSE
      b=CHR(bgcolour)
    ENDIF
  ELSE ' colour 0 => translate to "permanent black" 0xF0/240 (also in UTF-8)
    b=CHR(195, 144)
  ENDIF
  PRINTCOLOR=CHR(27)+"p"+f+b
END FUNCTION

SUB rgcSETPIXEL(BYVAL vout AS BYTE PTR, BYVAL colour AS UBYTE)
  DIM AS INTEGER r,g,b
  ' draws a pixel of a console character, performing effects if necessary
  IF colour<240 THEN ' ordinary colour
    *(vout)=colour
  ELSE ' effect colour (0xFx)
    SELECT CASE colour
      CASE 240:
        *vout=0 ' to the console 0xF0 is always black
      CASE 253: ' 0xFD - "relenting" colour, replaced by 0xFE
        *vout=253
      CASE 254: ' 0xFE - transparent colour, only replacing FX-colours
        IF *vout>=240 THEN
          *vout=254
        ENDIF
      CASE 255: ' 0xFF reverse transparent; inverts destination colour (a bit like XOR)
        colour=*vout
        IF colour<216 THEN ' console palette range
          ' invert RGB-based
          colour=215-colour
        ELSE ' user-defined colour, FX-colours are ignored in case
          ' swap colours from 216 to 239
          colour=216+(239-colour)
        ENDIF
      CASE 251: ' 0xFB - darkens ground by 33%
        colour=*vout
        IF colour<216 THEN ' only working with the RGB palette section
          r=colour MOD 6
          g=(colour/6) MOD 6
          b=colour/36
          r-=2
          g-=2
          b-=2
          IF r<0 THEN r=0
          IF g<0 THEN g=0
          IF b<0 THEN b=0
          colour=b*36+g*6+r
          *vout=colour
        ENDIF
      CASE 252: ' 0xFC - lightens ground by 33%
        colour=*vout
        IF colour<216 THEN ' only working with the RGB palette section
          r=colour MOD 6
          g=(colour/6) MOD 6
          b=colour/36
          r+=2
          g+=2
          b+=2
          IF r>5 THEN r=5
          IF g>5 THEN g=5
          IF b>5 THEN b=5
          colour=b*36+g*6+r
          *vout=colour
        ENDIF
      CASE ELSE
        ' reserved effect colours, don't do anything
    END SELECT
  ENDIF
END SUB

SUB rgcREFRESH
  ' puts out the whole console content, based on its char buffer
  DIM AS UBYTE PTR vout, cdp
  DIM AS INTEGER x, y, lo=0, ls, i, j
  DIM AS ULONG char
  DIM AS UBYTE cfl, fgcolour, bgcolour

  vout=SCREENPTR
  SCREENINFO lo, y
  lo-=8 ' offset to next line (after printing 8 pixels)

  FOR y=0 TO 29
    FOR x=0 TO 79
      IF rgccell(x,y).flags THEN ' character to print
        char=rgccell(x,y).char
        IF rgccell(x,y).flags AND 128 THEN
          ' inverse output
          fgcolour=rgccell(x,y).bgcolour
          bgcolour=rgccell(x,y).fgcolour
        ELSE
          fgcolour=rgccell(x,y).fgcolour
          bgcolour=rgccell(x,y).bgcolour
        ENDIF
        i=char SHR 7
        IF rgcfontpageoffset(i)=0 THEN ' "codepage" not existing => unknown char
          i=0
          char=0
        ENDIF
        ' get char font data start
        ls=rgcfontlinesize(i)
        cdp=@rgcfont(0)+rgcfontpageoffset(i)+(char AND 127)
        i=16
        WHILE i
          i-=1
          cfl=PEEK(UBYTE, cdp)
          j=8
          WHILE j
            j-=1
            IF (cfl AND 1) THEN ' foreground pixel
              IF fgcolour<240 THEN
                *vout=fgcolour
              ELSE
                rgcSETPIXEL(vout, fgcolour)
              ENDIF
            ELSE ' background pixel
              IF bgcolour<240 THEN
                *vout=bgcolour
              ELSE
                rgcSETPIXEL(vout, bgcolour)
              ENDIF
            ENDIF
            cfl SHR=1
            vout+=1
          WEND
          cdp+=ls
          vout+=lo
        WEND
        vout-=lo*16+120 ' minus 16 pixel lines on screen & 15*8 => 8 bytes/pixels right
      ELSE
        vout+=8 ' directly 8 pixels right
      ENDIF
    NEXT
    vout+=lo*16-512 ' to begin of next console line in screen memory
  NEXT
END SUB

SUB rgcCLS
  ' "Clear Screen" resp. "Clear Home"
  ' empties whole (console) screen, overwriting with current background colour
  ' deleting all characters in the screen buffer
  ' performs a "HOME"(AKA "Pos 1") action for the text cursor
  ' but doesn't change any settings though => may change the screen colour
  DIM AS UBYTE PTR vout
  DIM AS INTEGER x, y, lo

  vout=SCREENPTR
  SCREENINFO x, y
  IF x>640 THEN lo=x-640
  ' set console area to background colour
  IF rgcbgcolour<240 THEN
    FOR y=0 TO 479
      FOR x=0 TO 639
        *vout=rgcbgcolour
        vout+=1
      NEXT
      vout+=lo
    NEXT
  ELSE ' background colour is special effect colour
    FOR y=0 TO 479
      FOR x=0 TO 639
        rgcSETPIXEL(vout, rgcbgcolour)
        vout+=1
      NEXT
      vout+=lo
    NEXT
  ENDIF
  FOR y=0 TO 29
    FOR x=0 TO 79
      rgccell(x,y).flags=0
      IF rgcfgcolour=240 THEN
        rgccell(x,y).fgcolour=0 ' "permanent black"
      ELSE
        rgccell(x,y).fgcolour=rgcfgcolour
      ENDIF
      IF rgcbgcolour=240 THEN
        rgccell(x,y).bgcolour=0 ' "permanent black"
      ELSE
        rgccell(x,y).bgcolour=rgcbgcolour
      ENDIF
      rgccell(x,y).char=0
    NEXT
  NEXT
END SUB

SUB rgcSCROLLUP(BYVAL lines AS INTEGER=1)
  DIM AS INTEGER x, y

  IF lines<2 THEN ' =>scroll 1 line
    ' copy cells one line higher
    FOR y=0 TO 28
      FOR x=0 TO 79
        rgccell(x,y)=rgccell(x, y+1)
      NEXT
    NEXT
    ' empty last(new) line with spaces
    FOR x=0 TO 79
      rgccell(x,29).flags=1
      rgccell(x,29).fgcolour=rgcfgcolour
      rgccell(x,29).bgcolour=rgcbgcolour
      rgccell(x,29).char=32
    NEXT
  ELSE
    IF lines<29 THEN
      ' copy cells by <lines> line higher
      FOR y=0 TO 29-lines
        FOR x=0 TO 79
          rgccell(x,y)=rgccell(x, y+lines)
        NEXT
      NEXT
      ' empty last(new) lines with spaces
      FOR y=lines TO 29
        FOR x=0 TO 79
          rgccell(x,y).flags=1
          rgccell(x,y).fgcolour=rgcfgcolour
          rgccell(x,y).bgcolour=rgcbgcolour
          rgccell(x,y).char=32
        NEXT
      NEXT
    ELSE ' for the joker who wants to "scroll" more lines than a screen has
      rgcCLS
      EXIT SUB
    ENDIF
  ENDIF
  rgcREFRESH
END SUB

SUB rgcSCROLLDOWN(BYVAL lines AS INTEGER=1)
  DIM AS INTEGER x, y

  IF lines<2 THEN ' =>scroll 1 line
    ' copy cells one line down
    FOR y=28 TO 0 STEP -1
      FOR x=0 TO 79
        rgccell(x,y+1)=rgccell(x, y)
      NEXT
    NEXT
    ' empty first(new) line with spaces
    FOR x=0 TO 79
      rgccell(x,0).flags=1
      rgccell(x,0).fgcolour=rgcfgcolour
      rgccell(x,0).bgcolour=rgcbgcolour
      rgccell(x,0).char=32
    NEXT
  ELSE
    IF lines<29 THEN
      ' copy cells by <lines> line higher
      FOR y=29-lines TO 0 STEP -1
        FOR x=0 TO 79
          rgccell(x,y+lines)=rgccell(x, y)
        NEXT
      NEXT
      ' empty first(new) lines with spaces
      FOR y=0 TO lines-1
        FOR x=0 TO 79
          rgccell(x,y).flags=1
          rgccell(x,y).fgcolour=rgcfgcolour
          rgccell(x,y).bgcolour=rgcbgcolour
          rgccell(x,y).char=32
        NEXT
      NEXT
    ELSE ' for the joker who wants to "scroll" more lines than a screen has
      rgcCLS
      EXIT SUB
    ENDIF
  ENDIF
  rgcREFRESH
END SUB

SUB rgcBELL
  ' bell sound (alert sound)
  ' #todo:
  BEEP
END SUB

SUB rgcTYPE
  ' type sound
  ' #todo:
END SUB

SUB rgcCARRIAGERETURN
  ' carriage return sound
  ' #todo:
END SUB

SUB rgcMOVEON(BYVAL cells AS INTEGER=1)
  ' moves console "cursor" 1 space (after putting out a char mainly)
  ' #todo: maybe 4 cells for double-sized emphasized characters (or maybe stupid)
  IF cells<1 OR cells>2 THEN EXIT SUB ' illegal value
  rgccurx+=cells
  ' line overflow check
  IF rgccurx>80 THEN ' more 2 or more chars over the limit
    IF (rgcflags AND 2) THEN
      ' line overflow (to beginning of next line)
      rgccurx-=80
      rgccury+=1
      IF rgccury>29 THEN
        IF (rgcflags AND 1) THEN ' scrolling enabled
          RGCSCROLLUP
        ENDIF
          rgccury=29 ' scrolling disabled - trapped in last line
      ENDIF
    ELSE ' line overflow is off
      IF (rgcflags AND 128) THEN
        rgcBELL
      ENDIF
      rgccurx=79
    ENDIF
  ENDIF
END SUB

SUB rgcPRINTCHAR(BYVAL char AS ULONG) ' puts out given char (graphic, without control code handling)
  DIM AS UBYTE PTR cd, vout
  DIM AS INTEGER i, x, y, cx, cy, scw, ls
  DIM AS UBYTE cl

  ' scrolling is only performed when another character is put out
  IF rgccurx>79 THEN
    IF (rgcflags AND 2) THEN
      ' line overflow (to beginning of next line)
      rgccurx-=80
      rgccury+=1
      IF rgccury>29 THEN
        IF (rgcflags AND 1) THEN ' scrolling enabled
          rgcSCROLLUP
        ENDIF
        rgccury=29
      ENDIF
    ELSE ' line overflow is off
      IF (rgcflags AND 128) THEN
        rgcBELL
      ENDIF
      rgccurx=79
    ENDIF
  ENDIF
  
  vout=SCREENPTR ' error source: should be currently in the virtual 640x480 screen
  SCREENINFO scw,y

  i=char SHR 7 ' "character page"
  IF rgcfontpageoffset(i) THEN
    cd=@rgcfont(0)+rgcfontpageoffset(i)+(char AND 127) ' #note: this only works for 1 byte broad chars
    ls=rgcfontlinesize(i)
  ELSE ' offset is 0 => "page" not existing, put out "unknown" char (code 0)
    cd=@rgcfont(0)+rgcfontpageoffset(0) ' #note: "codepage" 0 must always exist
    ls=rgcfontlinesize(0)
    char=0
  ENDIF

  ' draw character
  vout+=rgccury*scw*16+rgccurx*8
  scw-=8
  IF rgccharflags AND 128 THEN ' inverted output
    IF rgcfgcolour>239 OR rgcbgcolour>240 THEN ' using effect colours
      ' slower like with normal colours
      FOR y=0 TO 15
        cl=PEEK(UBYTE, cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            rgcSETPIXEL(vout, rgcbgcolour)
          ELSE
            rgcSETPIXEL(vout, rgcfgcolour)
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ELSE ' normal colours, replacing background
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            *(vout)=rgcbgcolour
          ELSE
            *(vout)=rgcfgcolour
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ENDIF
  ELSE ' non-inverted output
    IF rgcfgcolour>239 OR rgcbgcolour>240 THEN ' using effect colours
      ' slower like with normal colours
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            rgcSETPIXEL(vout, rgcfgcolour)
          ELSE
            rgcSETPIXEL(vout, rgcbgcolour)
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ELSE ' normal colours, replacing background
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            *(vout)=rgcfgcolour
          ELSE
            *(vout)=rgcbgcolour
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ENDIF
  ENDIF

  ' store char in console buffer
  ' #todo:
  rgccell(rgccurx,rgccury).flags=rgccharflags
  rgccell(rgccurx,rgccury).fgcolour=rgcfgcolour
  rgccell(rgccurx,rgccury).bgcolour=rgcbgcolour
  rgccell(rgccurx,rgccury).char=char

  ' move 1 char #todo: double size characters => 2 cells
  rgcMOVEON
END SUB

SUB rgPRINT(BYVAL outputstring AS STRING)
  DIM AS INTEGER strend=LEN(outputstring), i, j
  DIM AS UBYTE c
  DIM AS ULONG uc

  IF strend=0 THEN EXIT SUB ' empty string
  strend-=1
  FOR i=0 TO strend
    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' maybe faster than MID$()...?
    IF c<32 THEN ' control code
      IF c THEN ' not 0
        IF c=27 THEN ' ESC (sequence)
          IF i<strend THEN
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
            IF c<128 THEN ' 7-bit / ASCII range
              SELECT CASE c
                CASE &H41: ' ESC+A inverse mode on
                  rgccharflags OR=128
                CASE &H61: ' ESC+a inverse mode off
                  rgccharflags AND=NOT 128
                CASE &H42: ' ESC+B bold on
                  rgccharflags OR=2
                CASE &H62: ' ESC+b bold off
                  rgccharflags AND=NOT 2
                CASE &H43: ' ESC+C canceled on
                  rgccharflags OR=16
                CASE &H63: ' ESC+c canceled off
                  rgccharflags AND=NOT 16
                CASE &H44: ' ESC+D disabled on
                  rgccharflags OR=32
                CASE &H64: ' ESC+d disabled off
                  rgccharflags AND=NOT 32
                CASE &H45: ' ESC+E emphasized/double width on
                  rgccharflags OR=1024
                CASE &H65: ' ESC+e normal width
                  rgccharflags AND=NOT 1024
                CASE &H46: ' ESC+F compressed mode on
                  rgccharflags OR=4096
                CASE &H47: ' ESC+G <bitmap data> graphic output (unsupported)
                CASE &H48: ' ESC+H double height on
                  rgccharflags OR=2048
                CASE &H68: ' ESC+h double height off
                  rgccharflags AND=NOT 2048
                CASE &H49: ' ESC+I italic on
                  rgccharflags OR=4
                CASE &H69: ' ESC+i italic off
                  rgccharflags AND=NOT 4
                CASE &H4A: ' ESC+J Form Feed / next page
                  ' on a screen, it's just like CLS - but we put a white paper in our printer ;-)...
                  rgcfgcolour=0
                  rgcbgcolour=215
                  rgcCLS
                CASE &H6A: ' ESC+j previous page
                  ' also nonsense for a screen
                  rgcCLS
                CASE &H4B: ' ESC+K erase rest of screen
                CASE &H6B: ' ESC+k erase rest of line
                CASE &H4D: ' ESC+M <c> "select codepage" (unsupported)
                  ' #todo: strip off character 'c' from string
                CASE &H4E: ' ESC+N all decorations&scalings off => normal text
                  rgccharflags=1
                CASE &H6E: ' ESC+n normal text size (Esc E/H/S/s off)
                  rgccharflags AND=NOT 7936
                CASE &H4F: ' ESC+O outlined on
                  rgccharflags OR=64
                CASE &H6F: ' ESC+o outlined off
                  rgccharflags AND=NOT 64
                CASE &H50: ' ESC+P place cursor <x> <y>
                  ' might use '127' to not change the coordinate, but "VT-2000" supports
                  ' in principle a maximum page size of 128x128 - from 1024x2048 up, overall
                  ' 3 page sizes with 128 characters per line - which is even not
                  ' entirely covered by ESC P
                  ' x and y must not be 0 nor above 127 though, they start with (1|1)
                  ' #note: internally, the coordinates are reaching from 0 to 79/29
                  IF (i+2)<strend THEN ' not yet string end reached
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c AND c<81 THEN ' other values are illegal
                      rgccurx=c-1
                    ENDIF
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c AND c<31 THEN ' other values are illegal
                      rgccury=c-1
                    ENDIF
                  ELSE
                    EXIT SUB
                  ENDIF
                CASE &H70: ' ESC+p set printing colours <fg> <bg>
                  ' => foreground & background colour
                  ' because the valid values range from 1 to 255, they may be also
                  ' 2 byte long UTF-8 encoded
                  ' 0 is illegal, you have to use 0xF0 instead chr(195,176)
                  IF (i+1)<strend THEN ' string not too short
                    ' foreground colour
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c THEN ' Nullbyte is illegal => you have to use 0xF0 / 240 CHR(195,176)
                      IF c<128 THEN ' colour below 128, easy
                        rgcfgcolour=c
                      ELSE ' UTF-8 encoded colour
                        IF c=194 THEN ' only those 2 values are valid
                          rgcfgcolour=128
                        ELSEIF c=195 THEN
                          rgcfgcolour=192
                        ELSE ' not ASCII nor UTF-8 (or valid)
                          EXIT SUB
                        ENDIF
                        i+=1
                        c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                        rgcfgcolour OR=(c AND 63)
                      ENDIF
                    ENDIF
                    i+=1
                    IF i>strend THEN EXIT SUB ' end of string
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    ' background colour
                    IF c THEN ' Nullbyte is illegal => you have to use 0xF0 / 240 CHR(195,176)
                      IF c<128 THEN ' colour below 128, easy
                        rgcbgcolour=c
                      ELSE ' UTF-8 encoded colour
                        IF c=194 THEN ' only those 2 values are valid
                          rgcbgcolour=128
                        ELSEIF c=195 THEN
                          rgcbgcolour=192
                        ELSE ' not ASCII nor UTF-8 (or valid)
                          EXIT SUB
                        ENDIF
                        i+=1
                        IF i>strend THEN EXIT SUB ' end of string
                        c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                        rgcbgcolour OR=(c AND 63)
                      ENDIF
                    ENDIF
                  ENDIF
                CASE &H51: ' ESC+Q <c> printing quality 0-127 (0 draft, 31 fast, 32 normal, 64+ highest)
                  ' #todo: strip off character 'c' from string
                CASE &H52: ' ESC+R proportional mode (reserved for)
                CASE &H72: ' ESC+r monospaced mode (default)
                CASE &H53: ' ESC+S superscript on
                  rgccharflags OR=512
                CASE &H73: ' ESC+s subscript on
                  rgccharflags OR=256
                CASE &H55: ' ESC+U underlined on
                  rgccharflags OR=8
                CASE &H75: ' ESC+U underlined off
                  rgccharflags AND=NOT 8
                CASE &H58: ' ESC+X show cursor (always)
                  rgcflags OR=32
                CASE &H78: ' ESC+x hide cursor
                  rgcflags AND=NOT 32
                CASE &H59: ' ESC+Y line overflow on
                  rgcflags OR=2
                CASE &H79: ' ESC+y line overflow off
                  rgcflags AND=NOT 2
                CASE &H5A: ' ESC+Z scrolling on
                  rgcflags OR=1
                CASE &H7A: ' ESC+z scrolling off
                  rgcflags AND=NOT 1
                CASE ELSE:
                  ' ignore illegal/unsupported ESC sequences, but put out remaining chars
                  i-=1
              END SELECT
            ELSE
              i-=1 ' illegal value => ignore "Mask" code, proceed
            ENDIF
          ELSE
            EXIT SUB
          ENDIF
        ELSE
          SELECT CASE c ' sorted by relevance
            CASE 13: ' newline (+scrolling)
              rgccurx=0
              rgccury+=1
              IF rgccury>29 THEN
                rgccury=29
                IF (rgcflags AND 1) THEN rgcSCROLLUP ' scrolling enabled
              ENDIF
            CASE 8: ' 8 tab right (ordinary tabulator, 8 char steps)
              IF rgccurx>78 THEN ' last column => new line
                rgccurx=79
                rgcMOVEON
              ELSE
                rgccurx+=8
                rgccurx AND=120
              ENDIF
            CASE 1: ' cursor up (+scrolling)
              IF rgccury THEN
                rgccury-=1
              ELSE
                IF (rgcflags AND 1) THEN rgcSCROLLDOWN
              ENDIF
            CASE 2: ' cursor down (+scrolling) "Line Feed" LF
              rgccury+=1
              IF rgccury>29 THEN
                rgccury=29
                rgcSCROLLUP
              ENDIF
            CASE 3: ' cursor left
              IF rgccurx THEN
                rgccurx-=1
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 4: ' cursor right
              IF rgccurx>78 THEN
                rgccurx=0
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 26: ' Mask
              ' puts out next char's graphic minus bit 5, e.g. a control code graphic
              IF i<strend THEN
                i+=1
                c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                ' #todo: perhaps add UTF-8 support
                IF c<128 THEN ' 7-bit / ASCII range
                  c AND=95
                  rgcPRINTCHAR(CAST(ULONG,c))
                ELSE
                  i-=1 ' illegal value => ignore "Mask" code, proceed
                ENDIF
              ELSE
                EXIT SUB
              ENDIF
            CASE 24: ' 4 tab right
              IF rgccurx>78 THEN ' last column => new line
                rgcMOVEON
              ELSE
                rgccurx+=4
                rgccurx AND=124
              ENDIF
            CASE 14: ' clear line CL (+ code 11)
            CASE 15: ' clear screen CLS (+ HOME = code 11+9 = code 16)
              rgcCLS
            CASE 16: ' HOME ("Pos 1")
              rgccurx=0
              rgccury=0
            CASE 19: ' Backspace
            CASE 20: ' Delete
            CASE 25: ' Overwrite
            CASE 28: ' redraw screen
            CASE 23: ' 4 tab left (previous tab position)
              IF rgccurx THEN
                rgccurx-=1
                rgccurx AND=124
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 7: ' 8 tab left (previous tab position)
              IF rgccurx THEN
                rgccurx-=1
                rgccurx AND=120
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 9: ' cursor to top
              rgccury=0
            CASE 10: ' cursor to bottom
              rgccury=29
            CASE 11: ' cursor to left border / "Carriage Return" CR
              rgccurx=0
            CASE 12: ' cursor to right border
              rgccurx=79
            CASE 29: ' carriage returning sound (scrolling)
              rgcCARRIAGERETURN
            CASE 30: ' typing sound
              rgcTYPE
            CASE 31: ' Bell / Alert, End Of Line (right paper edge) signal/sound
              rgcBELL
            CASE ELSE: ' unsupported & illegal control codes => ignored
              ' 5, 6, 21, 22 vertical "tabulators" => useless
          END SELECT
        ENDIF
      ELSE ' c is 0
        EXIT SUB ' Nullbyte => break output
      ENDIF
    ELSE ' normal character
      IF c>127 THEN ' UTF-8 char
        IF (c AND 192)=192 THEN ' UTF-8 start character
          IF (c AND 224)=192 THEN ' "110..." => 2 byte character
            uc=(c AND 63) SHL 6
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i)
            IF (c AND 192)=128 THEN ' okay
              uc+=(c AND 63)
              rgcPRINTCHAR(uc)
            ELSE
              EXIT SUB ' illegal UTF-8 code #todo:
            ENDIF
          ELSEIF (c AND 240)=224 THEN ' "1110..." => 3 byte character
            uc=(c AND 15) SHL 12
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i)
            IF (c AND 192)=128 THEN ' okay
              uc+=(c AND 63) SHL 6
              i+=1
              c=PEEK(UBYTE, STRPTR(outputstring)+i)
              IF (c AND 192)=128 THEN ' okay
                uc+=(c AND 63)
                rgcPRINTCHAR(uc)
              ELSE
                EXIT SUB ' illegal UTF-8 code #todo:
              ENDIF
            ELSE
              EXIT SUB ' illegal UTF-8 code #todo:
            ENDIF
          ELSEIF (c AND 248)=240 THEN ' "11110..." => 4 byte character
            uc=(c AND 63) SHL 6
          ELSE ' 5 byte character or more => illegal
            EXIT SUB ' #todo: maybe another handling / reaction; error code
          ENDIF
        ELSE
          ' not a UTF-8 char #todo: ? - could be only some binary data or ASCII char, but...
        ENDIF
      ELSE ' ASCII char
        uc=c
        rgcPRINTCHAR(uc)
      ENDIF
    ENDIF
  NEXT
END SUB

SUB rgINPUT(BYVAL prompt AS STRING, BYREF inputstr AS STRING, BYVAL limit AS INTEGER=72)
  ' limit sind Zeichen, nicht Bytes
  ' stringeingabe bis bildschirmende, overflow (intern) off, inverser cursor
  ' selbstgebasteltes Scrolling in "Eingabebereich", Pfeile <- ->
  ' tastatureingabe ASCII + UTF8
  ' nur Zeichen >31, Sondertasten filtern, ggf. zustzlich gefiltert nach existenten Zeichen
  ' bersetzung von Ziffernblock in Ziffern, ggf. nur mit Num Lock
  ' Enter+Return besttigen die Eingabe
  ' cursor <-> STRG+Cursor springt von Wort zu Wort, Pos1 und Ende
  ' SCREENSHOW regelmig fr Mauspfeil, blockiert alle anderen Eingaben
  ' Abbruch ("") mit ESC, rechtem Mausklick und linkem Mausklick deutlich auerhalb Eingabebereich
  ' Cursorplatzierung mit Maus
  ' Spter auch Markieren mit Shift+Cursortasten und Maus => Copy&Paste Funktion
END SUB
